package org.dasein.cloud.jclouds.gogrid;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Locale;

import org.apache.log4j.Logger;
import org.dasein.cloud.AsynchronousTask;
import org.dasein.cloud.CloudException;
import org.dasein.cloud.InternalException;
import org.dasein.cloud.OperationNotSupportedException;
import org.dasein.cloud.services.image.ImageServices;
import org.dasein.cloud.services.image.ServerImage;
import org.dasein.cloud.services.server.Platform;
import org.dasein.cloud.services.server.ServerSize;
import org.dasein.cloud.services.storage.CloudStoreObject;
import org.jclouds.gogrid.GoGridAsyncClient;
import org.jclouds.gogrid.GoGridClient;
import org.jclouds.gogrid.options.GetImageListOptions;
import org.jclouds.rest.RestContext;

public class GoGridImage implements ImageServices {
    static private final Logger logger = Logger.getLogger(GoGridImage.class);
    
    private GoGrid provider;
    
    GoGridImage(GoGrid gogrid) {
        provider = gogrid;
    }

    @Override
    public String create(String location) throws InternalException, CloudException {
        throw new OperationNotSupportedException("GoGrid does not build images in this manner.");
    }

    @Override
    public void delete(String imageId) throws InternalException, CloudException {
        throw new OperationNotSupportedException("GoGrid does not allow you to delete images.");
    }

    @Override
    public Collection<ServerImage> find(String executableByAccountNumber, String keyword, Platform platform, String architecture) throws InternalException, CloudException {
        ArrayList<ServerImage> list = new ArrayList<ServerImage>();
        
        for( ServerImage image : list(provider.getContext().getAccountNumber()) ) {
            if( platform != null && !platform.equals(image.getPlatform()) ) {
                continue;
            }
            if( keyword != null && image.getName().indexOf(keyword) == -1 ) {
                continue;
            }
            list.add(image);
        }
        return list;
    }

    @Override
    public ServerImage getImage(String imageId) throws InternalException, CloudException {
        return toImage(getGoGridImage(imageId));
    }

    private org.jclouds.gogrid.domain.ServerImage getGoGridImage(String imageId) throws CloudException {
        RestContext<GoGridClient,GoGridAsyncClient> ctx = provider.getCloudClient();
        
        try {
            for( org.jclouds.gogrid.domain.ServerImage image : ctx.getApi().getImageServices().getImagesById(Long.parseLong(imageId)) ) {
                return image;
            }
        }
        catch( RuntimeException e ) {
            String msg = e.getMessage();
            
            if( msg != null && msg.contains("No object found") ) {
                return null;
            }
            logger.error("Failed to lookup image " + imageId + ": " + e.getMessage());
            e.printStackTrace();
            throw new CloudException("Failed to lookup image " + imageId + ": " + e.getMessage());
        }
        finally {
            ctx.close();
        }
        return null;
    }
    
    @Override
    public CloudStoreObject getManifest(String location) throws CloudException, InternalException {
        // GoGrid does not store images on cloud storage
        return null;
    }

    @Override
    public String getProviderTermForImage(Locale locale) {
        return "image";
    }

    @Override
    public Collection<String> getShares(String imageId) throws CloudException, InternalException {
        return new ArrayList<String>();
    }

    @Override
    public AsynchronousTask<ServerImage> imageServer(String serverId, String forAccountAccessKey, String directory, String name) throws InternalException, CloudException {
        throw new OperationNotSupportedException();
    }

    @Override
    public boolean isPublic(String imageId) throws InternalException, CloudException {
        return true;
    }

    @Override
    public boolean isSupportsCustomImages() throws InternalException, CloudException {
        return false;
    }

    @Override
    public boolean isSupportsSharing() throws InternalException, CloudException {
        return false;
    }

    @Override
    public Collection<ServerImage> list(String accountNumber) throws InternalException, CloudException {
        RestContext<GoGridClient,GoGridAsyncClient> ctx = provider.getCloudClient();
        
        try {
            ArrayList<ServerImage> images = new ArrayList<ServerImage>();
            
            for( org.jclouds.gogrid.domain.ServerImage image : ctx.getApi().getImageServices().getImageList(new GetImageListOptions[] {}) ) {
                ServerImage img = toImage(image);
                
                if( img != null ) {
                    images.add(img);
                }
            }
            return images;
        }
        catch( RuntimeException e ) {
            logger.error("Failed to list images: " + e.getMessage());
            e.printStackTrace();
            throw new CloudException("Failed to list images: " + e.getMessage());
        }
        finally {
            ctx.close();
        }
    }

    @Override
    public void share(String imageId, String withAccountNumber, boolean allowed) throws InternalException, CloudException {
        throw new OperationNotSupportedException("GoGrid does not support image sharing.");
    }
    
    private ServerImage toImage(org.jclouds.gogrid.domain.ServerImage goGridImage) throws InternalException, CloudException {
        if( goGridImage == null ) {
            return null;
        }
        ServerImage image = new ServerImage();
        
        image.setActive(goGridImage.isActive());
        image.setArchitecture(provider.toArchitecture(goGridImage.getOs(), goGridImage.getDescription()));
        image.setAvailable(true);
        image.setDescription(goGridImage.getDescription());
        image.setName(goGridImage.getFriendlyName());
        image.setOwnerAccountId(provider.getContext().getAccountNumber());
        image.setPersistent(true);
        image.setPlatform(provider.toPlatform(goGridImage.getOs()));
        image.setProviderServerImageId(String.valueOf(goGridImage.getId()));
        image.setPublicImage(true);

        Collection<ServerSize> sizes = provider.getServerServices().getSupportedSizes(image.getArchitecture());
        ServerSize[] tmp = new ServerSize[sizes.size()];
        
        tmp = sizes.toArray(tmp);
        image.setSize(tmp);
        image.setType("standard");
        return image;
    }
}
