package org.dasein.cloud.jclouds.gogrid;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Locale;
import java.util.Set;

import org.apache.log4j.Logger;
import org.dasein.cloud.CloudException;
import org.dasein.cloud.InternalException;
import org.dasein.cloud.OperationNotSupportedException;
import org.dasein.cloud.services.server.Architecture;
import org.dasein.cloud.services.server.Server;
import org.dasein.cloud.services.server.ServerPersistence;
import org.dasein.cloud.services.server.ServerServices;
import org.dasein.cloud.services.server.ServerSize;
import org.dasein.cloud.services.server.ServerState;
import org.dasein.cloud.services.server.ServerStatistics;
import org.jclouds.gogrid.GoGridAsyncClient;
import org.jclouds.gogrid.GoGridClient;
import org.jclouds.gogrid.GoGridResponseException;
import org.jclouds.gogrid.domain.Ip;
import org.jclouds.gogrid.domain.Option;
import org.jclouds.gogrid.domain.PowerCommand;
import org.jclouds.gogrid.options.AddServerOptions;
import org.jclouds.gogrid.options.GetServerListOptions;
import org.jclouds.rest.RestContext;

public class GoGridServer implements ServerServices {
    static private final Logger logger = Logger.getLogger(GoGridServer.class);
    
    private final GoGrid provider;
    
    GoGridServer(GoGrid gogrid) {
        provider = gogrid;
    }
    
    @Override
    public void boot(String serverId) throws InternalException, CloudException {
        RestContext<GoGridClient,GoGridAsyncClient> ctx = provider.getCloudClient();

        try {   
            ctx.getApi().getServerServices().power(serverId, PowerCommand.START);
        }
        catch( RuntimeException e ) {
            logger.error("Failed to power on server " + serverId + ": " + e.getMessage());
            e.printStackTrace();
            throw new CloudException("Failed to power on server " + serverId + ": " + e.getMessage());
        }
        finally {
            ctx.close();
        }
    }

    public Server clone(String serverId, String intoDcId, String name, String description, boolean powerOn, String ... firewallIds) throws InternalException, CloudException {
        throw new OperationNotSupportedException("GoGrid does not support server cloning.");
    }
    
    @Override
    public Server define(String imageId, ServerSize size, String dataCenterId, String name, boolean withAnalytics, String ... firewalls) throws InternalException, CloudException {
        RestContext<GoGridClient,GoGridAsyncClient> ctx = provider.getCloudClient();
        Set<Ip> availableIps = ctx.getApi().getIpServices().getUnassignedIpList();
        
        try {
            ArrayList<Long> ids = new ArrayList<Long>();
            boolean waiting = true;
            Ip address = null;
            
            // GoGrid doesn't return a server ID immediately on addServer, so we have to figure it out
            while( waiting ) {
                waiting = false;
                try {
                    for( org.jclouds.gogrid.domain.Server server : ctx.getApi().getServerServices().getServersByName(name) ) {
                        if( server.getId() == 0L ) {
                            waiting = true;
                            ids.clear();
                            break;
                        }
                        ids.add(server.getId());
                    }
                }
                catch( GoGridResponseException ignore ) {
                    // ignore this
                }
                if( waiting ) {
                    try { Thread.sleep(10000L); }
                    catch( InterruptedException e ) { }
                }
            }
            for( Ip ip : availableIps ) {
                if( ip.isPublic() ) {
                    address = ip;
                }
            }
            if( address == null ) {
                throw new CloudException("No more addresses are left in this account.");
            }
            org.jclouds.gogrid.domain.Server server;
            
            try {   
                AddServerOptions opts = new AddServerOptions();
                
                opts.withDescription(name);
                server = ctx.getApi().getServerServices().addServer(name, imageId, size.getSizeId(), address.getIp(), opts);
            }
            catch( GoGridResponseException e ) {
                String msg = e.getMessage();
                
                if( msg != null && msg.contains("IP already in use") ) {
                    return define(imageId, size, dataCenterId, name, withAnalytics, firewalls);
                }
                throw new CloudException(e.getMessage());
            }
            if( server == null ) {
                throw new CloudException("Failed to create server " + name);
            }
            server = null;
            while( server == null ) {
                try {
                    for( org.jclouds.gogrid.domain.Server s : ctx.getApi().getServerServices().getServersByName(name) ) {
                        long id = s.getId();
                        
                        if( id != 0L && !ids.contains(id) ) {
                            server = s;
                            break;
                        }
                    }
                }
                catch( GoGridResponseException ignore ) {
                    // waiting
                }
                if( server == null ) {
                    try { Thread.sleep(10000L); }
                    catch( InterruptedException e ) { }
                }
            }
            return toServer(server);
        }
        catch( RuntimeException e ) {
            throw new CloudException("Failed to define server: " + e.getMessage());
        }
        catch( Error e ) {
            throw new InternalException("Failed to define server: " + e.getMessage());
        }
        finally {
            ctx.close();
        }
    }

    @Override
    public String getConsoleOutput(String serverId, long since) throws InternalException, CloudException {
        return "";
    }

    @Override
    public Collection<String> getFirewalls(String serverId) throws InternalException, CloudException {
        return new ArrayList<String>();
    }

    @Override
    public String getProviderTermForServer(Locale locale) {
        return "server";
    }

    @Override
    public Server getServer(String serverId) throws InternalException, CloudException {
        return toServer(getGoGridServer(serverId));
    }

    private org.jclouds.gogrid.domain.Server getGoGridServer(String serverId) throws CloudException {
        RestContext<GoGridClient,GoGridAsyncClient> ctx = provider.getCloudClient();
        long id;
        
        try {
            id = Long.parseLong(serverId);
        }
        catch( NumberFormatException e ) {
            return null;
        }
        try {
            for( org.jclouds.gogrid.domain.Server server : ctx.getApi().getServerServices().getServersById(id) ) {
                return server;
            }        
        }
        catch( RuntimeException e ) {
            String msg = e.getMessage();
            
            if( msg.contains("No object found") ) {
                return null;
            }
            logger.error("Failed to find server " + serverId + ": " + e.getMessage());
            e.printStackTrace();
            throw new CloudException("Failed to find server " + serverId + ": " + e.getMessage());
        }
        finally {
            ctx.close();
        }
        return null;
    }
    
    @Override
    public ServerPersistence getServerPersistence() throws InternalException, CloudException {
        return ServerPersistence.ALL;
    }

    @Override
    public ServerStatistics getServerStatistics(String serverId, long begin, long end) throws InternalException, CloudException {
        return new ServerStatistics();
    }

    @Override
    public Collection<ServerStatistics> getServerStatisticsForPeriod(String serverId, long start, long end) throws InternalException, CloudException {
        return new ArrayList<ServerStatistics>();
    }

    
    @Override
    public Collection<ServerSize> getSupportedSizes(Architecture architecture) throws InternalException, CloudException {
        RestContext<GoGridClient,GoGridAsyncClient> ctx = provider.getCloudClient();

        try {
            Set<Option> sizes = ctx.getApi().getServerServices().getRamSizes();
            ArrayList<ServerSize> list = new ArrayList<ServerSize>();
            
            for( Option size : sizes ) {
                list.add(new ServerSize(String.valueOf(size.getId()), size.getName()));
            }
            return list;
        }
        catch( Throwable t ) {
            throw new CloudException(t);
        }
        finally {
            ctx.close();
        }
    }

    @Override
    public boolean isSupportsCloning() {
        return false;
    }
    
    @Override
    public Server launch(String imageId, ServerSize size, String dataCenterId, String serverName, String withKey, String inVlanId, boolean withMonitoring, String... firewalls) throws InternalException, CloudException {
        Server server = define(imageId, size, dataCenterId, serverName, withMonitoring, firewalls);
        
        //boot(server.getProviderServerId());
        return server;
    }

    @Override
    public Collection<Server> list() throws InternalException, CloudException {
        RestContext<GoGridClient,GoGridAsyncClient> ctx = provider.getCloudClient();
        ArrayList<Server> servers = new ArrayList<Server>();
        Set<org.jclouds.gogrid.domain.Server> list;
        
        try {
            list = ctx.getApi().getServerServices().getServerList(GetServerListOptions.NONE);
        }
        catch( RuntimeException e ) {
            logger.error("Failed to list servers: " + e.getMessage());
            e.printStackTrace();
            throw new CloudException("Failed to list servers: " + e.getMessage());
        }
        finally {
            ctx.close();
        }
        for( org.jclouds.gogrid.domain.Server server : list ) {
            servers.add(toServer(server));
        }
        return servers;
    }

    @Override
    public void monitor(String serverId) throws InternalException, CloudException {
        // NO-OP
    }

    @Override
    public void pause(String serverId) throws InternalException, CloudException {
        RestContext<GoGridClient,GoGridAsyncClient> ctx = provider.getCloudClient();

        try {
            ctx.getApi().getServerServices().power(serverId, PowerCommand.STOP);
        }
        catch( RuntimeException e ) {
            logger.error("Failed to pause server " + serverId + ": " + e.getMessage());
            e.printStackTrace();
            throw new CloudException("Failed to pause server " + serverId + ": " + e.getMessage());
        }
        finally {
            ctx.close();
        }
    }

    @Override
    public void reboot(String serverId) throws CloudException, InternalException {
        RestContext<GoGridClient,GoGridAsyncClient> ctx = provider.getCloudClient();

        try {
            ctx.getApi().getServerServices().power(serverId, PowerCommand.RESTART);
        }
        catch( RuntimeException e ) {
            logger.error("Failed to reboot server " + serverId + ": " + e.getMessage());
            e.printStackTrace();
            throw new CloudException("Failed to reboot server " + serverId + ": " + e.getMessage());
        }
        finally {
            ctx.close();
        }
    }

    @Override
    public void stop(String serverId) throws InternalException, CloudException {
        RestContext<GoGridClient,GoGridAsyncClient> ctx = provider.getCloudClient();

        try {
            ctx.getApi().getServerServices().deleteById(Long.parseLong(serverId));        
        }
        catch( RuntimeException e ) {
            logger.error("Failed to find delete " + serverId + ": " + e.getMessage());
            e.printStackTrace();
            throw new CloudException("Failed to find delete " + serverId + ": " + e.getMessage());
        }
        finally {
            ctx.close();
        }
    }

    @Override
    public void unmonitor(String serverId) throws InternalException, CloudException {
        // NO-OP
    }
    
    private ServerSize toSize(Option ram) {
        return new ServerSize(String.valueOf(ram.getId()), ram.getName());
    }
    

    private ServerState toState(org.jclouds.gogrid.domain.ServerState serverState) {
       //   ON, STARTING, OFF, STOPPING, RESTARTING, SAVING, RESTORING, UPDATING, UNRECOGNIZED;
        if(serverState ==  null|| serverState ==  org.jclouds.gogrid.domain.ServerState.UNRECOGNIZED || serverState == org.jclouds.gogrid.domain.ServerState.STARTING ) { 
            return ServerState.PENDING;
        }
        switch( serverState ) {
            case ON: return ServerState.RUNNING; 
            case OFF: return ServerState.PAUSED;
            case STOPPING: return ServerState.STOPPING;
            case RESTARTING: return ServerState.REBOOTING;
            case SAVING: return ServerState.REBOOTING;
            case RESTORING: return ServerState.PENDING;
        }
        return ServerState.RUNNING;
    }
    
    private Server toServer(org.jclouds.gogrid.domain.Server fromGoGrid) {
        if( fromGoGrid != null && fromGoGrid.getId() > 0 ) {
            org.jclouds.gogrid.domain.ServerImage image = fromGoGrid.getImage();
            Server server = new Server();
            
            server.setName(fromGoGrid.getName());
            server.setProviderServerId(String.valueOf(fromGoGrid.getId()));
            server.setImageId(String.valueOf(image.getId()));
            server.setArchitecture(provider.toArchitecture(image.getOs(), fromGoGrid.getDescription()));
            Ip ip = fromGoGrid.getIp();
            
            if( ip != null ) {
                if( ip.isPublic() ) {
                    server.setPublicIpAddresses(ip.getIp());
                    server.setReservedAddressId(String.valueOf(ip.getId()));
                }
                else {
                    server.setPrivateIpAddresses(ip.getIp());
                }
            }
            server.setCreateTime(0L);
            server.setCurrentState(toState(fromGoGrid.getState()));
            server.setDataCenterId(String.valueOf(fromGoGrid.getDatacenter().getId()));
            server.setDescription(fromGoGrid.getDescription());
            server.setPersistent(true);
            server.setPlatform(provider.toPlatform(fromGoGrid.getOs()));
            server.setRegionId("1");
            server.setSize(toSize(fromGoGrid.getRam()).getSizeId());
            server.setStartTime(0L);
            server.setStopTime(-1L);
            server.setTerminationTime(-1L);
            return server;
        }
        return null;
    }
}
